#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include "mylib.h"

#if defined DEBUG
#define DEBUGMEM(func,size)  printf(" %s %d bytes\n",func,size)
#else
#define DEBUGMEM(func,size)
#endif

#define abs(i) ((i<0)?-i:i)
#define max2(i,j) ((i>j)?i:j)
#define max3(i,j,k) ((i>j)?max2(i,k):max2(j,k))
#define max4(i,j,k,l) ((i>j)?max3(i,k,l):max3(j,k,l))

/*
 *  function realloc with error checking
 */
void *myrealloc(void *ptr, size_t size) {
  void *p = realloc(ptr,size);
  DEBUGMEM("realloc",size);
  /* error? */
  if (p==NULL && size!=0) {
    fprintf(stderr,"Allocation memory error trying to reallocate %d bytes\n.",size);
    exit(1);				/* abort program */
  }
  return p;
}

/*
 *  function malloc with error checking
 */
void *mymalloc(size_t size) {
  void *p = malloc(size);
  DEBUGMEM("malloc",size);
  /* error? */
  if (p==NULL && size!=0) {
    fprintf(stderr,"Allocation memory error trying to allocate %d bytes\n.",size);
    exit(1);				/* abort program */
  }
  return p;
}

/*
 *  Read sequence from file f with FASTA format and store it in seq.
 *  Alloc enough memory to store the sequence.
 *  Returns the size of the sequence or 0 on error. 
 */
int readsequence(FILE *file, char**seq, char **name){
  char base;
  int sizeseq, sizename;
  const int chunk=128;

  sizeseq = 0;
  *seq = NULL;
  sizename = 0;
  *name = NULL;

  //waitgreater:
  for ( base = (char) fgetc(file);
	base != '>';
	base = (char) fgetc(file) )
    if(base == EOF)
      goto error;
  
      
  //waitendofline:
  for ( base = (char) fgetc(file);
	base != '\n';
	base = (char) fgetc(file) ) {
    if(base == EOF)
      goto error;
    if (sizename % chunk == 0)
      *name = (char *) myrealloc( *name, (sizename+chunk)*sizeof(char) );
    (*name)[sizename] = base;
    sizename++;
  }
  *name = (char *) myrealloc( *name, (sizename+1)*sizeof(char) );
  (*name)[sizename] = '\0';
  
  
  //readseq:
  for ( base = toupper((char) fgetc(file));
	base != '>' && base != EOF;
	base = toupper((char) fgetc(file)) )
    if ( base == 'A' || base == 'C' || base == 'T'
	 || base == 'G' || base=='N') {  /* N means any base */
      if (sizeseq % chunk == 0)
	*seq = (char *) myrealloc( *seq, (sizeseq+chunk)*sizeof(char) );
      (*seq)[sizeseq] = base;
      sizeseq++;
    }
  *seq = (char *) myrealloc( *seq, (sizeseq+1)*sizeof(char) );
  (*seq)[sizeseq] = '\0';
  
  if (base!=EOF) ungetc(base,file);	/* to allow reading multifasta files */
  return sizeseq;
  
 error:
  free(*name);
  free(*seq);
  *name=NULL;
  *seq=NULL;
  return 0;
}

/*
 *  Dynamically allocate a matrix of integers with n rows and m columns
 */
int	**allocmatrix( int n, int m) {
  int i;
  int **mat;
  /* alloc n rows */
  mat = (int **) mymalloc( (n+1)*sizeof(int *) );
  /* each of it */
  for (i=0; i<n; i++) {
    /* with m columns */
    mat[i] = (int *) mymalloc( m*sizeof(int) );
  }
  mat[i]=NULL;				/* an extra row to mark the end */
  return mat;
}
  


/*
 *  Deallocate a matrix of integers with n rows and m columns
 */
void	freematrix( int **mat ) {
  int i;
  if (mat!=NULL) {
    for(i=0; mat[i]!=NULL; i++)
      free(mat[i]);
    free(mat);
  }
}

/*
 * Returns in Seq a random DNA sequence of length Len 
 */

char *RandomSeq(char *Seq, int Len) {
  int i;
  char alpha[]="ACTG";
  for (i=0;i<Len;i++)
    Seq[i] = alpha[ (int)(4.0*random()/(RAND_MAX+1.0)) ];
  Seq[i] = '\0';
  return Seq;
}


void SeedRandomSeq(unsigned int Seed){
  if (!Seed) {
    struct timeval tv;
    gettimeofday(&tv,NULL);
    Seed=(unsigned int)(tv.tv_sec ^ tv.tv_usec);
  }
  srandom(Seed);
}

/*
 * Returns in RevComp a the reverse complementation of Sequence Seq of length Len
 */
char *ReverseComplement(char *RevComp, char *Seq, int Len) {
  int i;
  for (i=0;i<Len;i++) {
    switch (toupper(Seq[Len-i-1])) {
    case 'A':
      RevComp[i] = 'T';
      break;
    case 'T':
      RevComp[i] = 'A';
      break;
    case 'C':
      RevComp[i] = 'G';
      break;
    case 'G':
      RevComp[i] = 'C';
      break;
    case 'N':
      RevComp[i] = 'N';
      break;
    default:
      RevComp[i] = toupper(Seq[Len-i-1]);
    }
  }
  RevComp[i] = '\0';
  return RevComp;
}

/*
 * Print Alignment of two sequences.
 *    Sequence of name name1 starts in position from1 and finishes in position to1.
 *    Sequence of name name2 starts in position from2 and finishes in position to2.
 *    The alignment is given by two rows align1 and align2 of length len and
 *       score is its punctuation.  Gaps are represented by ' ' or '-'.
 *    At most, maxcols columns of alignment per line are printed.  It defaults
 *       to 60, if maxcols <= 0.
 */
void PrintAlignment( char *name1, int from1, int to1, char *align1,
		     char *name2, int from2, int to2, char *align2,
		     int len, int maxcols, int score ) {
  int i,j,lenpos,inc1,inc2;
  int iden=0, gaps=0;
#define isidentity(i) (align1[i]==align2[i] && (tolower(align1[i])=='a'||tolower(align1[i])=='c'||tolower(align1[i])=='t'||tolower(align1[i])=='g'))
#define isgap(a) (a==' '||a=='-')
  
  for (i=0; i<len; i++){
    if (isidentity(i))
      iden++;
    if (isgap(align1[i])) 
      gaps++;
    if (isgap(align2[i])) 
      gaps++;
  }

  j=i=max2( abs(to1-from1), abs(to2-from2) )+ 1;
  for (lenpos=0; i; i/=10)
    lenpos++;
    
  printf( "Sequence 1: %s\n"
	  "Sequence 2: %s\n\n"
	  "Score:      %d\n"
	  "MaxLength:  %*d\n"
	  "Identities: %*d (%d%%)\n"
	  "Gaps:       %*d (%d%%)\n\n",
	  name1, name2, score,
	  lenpos, j,
	  lenpos, iden, (int)(100.0*iden/j),
	  lenpos, gaps, (int)(100.0*gaps/j)
	  );
  
  i=max4(from1,to1,from2,to2);
  for (lenpos=0; i; i/=10)
    lenpos++;
  inc1=(to1>from1)?+1:-1;
  inc2=(to2>from2)?+1:-1;
  if (maxcols<=0)
    maxcols=60;

  for (j=i=0; i<len; ) {
    printf("\n");

    printf("Seq1: %*d ", lenpos, from1);
    do {
      if (!isgap(align1[i]))
	from1+=inc1;
      printf("%c",tolower(align1[i++]));
    } while (i<len && i-j!=maxcols);
    printf(" %*d\n", lenpos, from1-inc1);

    i=j;
    
    printf("      %*s ", lenpos, "");
    do {
      printf("%c",isidentity(i)?'|':' ');
      i++;
    } while (i<len && i-j!=maxcols);
    printf(" %*s\n", lenpos, "");

    i=j;

    printf("Seq2: %*d ", lenpos, from2);
    do {
      if (!isgap(align2[i]))
	from2+=inc2;
      printf("%c",tolower(align2[i++]));
    } while (i<len && i-j!=maxcols);
    printf(" %*d\n", lenpos, from2-inc2);

    printf("\n");
    j=i;
  }

  if (to1!=from1-inc1)
    printf("Warning: end of Seq1 is %d (not %d)\n", from1-inc1, to1);
  if (to2!=from2-inc2)
    printf("Warning: end of Seq2 is %d (not %d)\n", from2-inc2, to2);
}
